from sanji.model_initiator import ModelInitiator
from voluptuous import Schema
from threading import Event
from six import moves


class ModelBatch(object):

    def __init__(self, model):
        self.enable = Event()
        self.model = model

    def __enter__(self):
        if self.enable.is_set():
            raise RuntimeError("Already in batch mode")
        self.enable.set()

    def __exit__(self, type, value, traceback):
        self.enable.clear()
        self.model.save_db()


class Model(object):

    def __init__(self, name, path, schema=None, model_cls=dict):
        self.model_cls = model_cls
        self.schema = schema

        if schema is not None and not isinstance(schema, Schema):
            raise TypeError("schema should be instance of voluptuous.Schema")

        if not issubclass(model_cls, dict):
            raise TypeError("model_cls should be derivative dict class")

        self.model = ModelInitiator(
            model_name=name,
            model_path=path
        )
        self._batch = ModelBatch(self.model)

    def batch(self):
        return self._batch

    def _cast_model(self, obj):
        return self.model_cls(obj)

    @property
    def maxId(self):
        """int: current max id of objects"""
        if len(self.model.db) == 0:
            return 0

        return max(map(lambda obj: obj["id"], self.model.db))

    def validation(self, instance):
        """Valid input instance is vaild or not
            Args:
                Object: input instance
            Returns:
                Object: Instance after vaildation or original instance
                        if schema is None
            Raises:
                Error: If vaildation failed
        """
        if self.schema is None:
            return instance

        return self.schema(instance)

    def add(self, obj):
        """Add a object
            Args:
                Object: Object will be added
            Returns:
                Object: Object with id
            Raises:
                TypeError: If add object is not a dict
                MultipleInvalid: If input object is invaild
        """
        if not isinstance(obj, dict):
            raise TypeError("Add object should be a dict object")
        obj = self.validation(obj)
        obj["id"] = self.maxId + 1
        obj = self._cast_model(obj)
        self.model.db.append(obj)

        if not self._batch.enable.is_set():
            self.model.save_db()
        return obj

    def get(self, id):
        """Get a object by id
            Args:
                id (int): Object id

            Returns:
                Object: Object with specified id
                None: If object not found
        """
        for obj in self.model.db:
            if obj["id"] == id:
                return self._cast_model(obj)

        return None

    def remove(self, id):
        """Remove a object by id
            Args:
                id (int): Object's id should be deleted
            Returns:
                len(int): affected rows
        """
        before_len = len(self.model.db)
        self.model.db = [t for t in self.model.db if t["id"] != id]
        if not self._batch.enable.is_set():
            self.model.save_db()
        return before_len - len(self.model.db)

    def removeAll(self):
        """Remove all objects
            Returns:
                len(int): affected rows
        """
        before_len = len(self.model.db)
        self.model.db = []
        if not self._batch.enable.is_set():
            self.model.save_db()
        return before_len - len(self.model.db)

    def update(self, id, newObj):
        """Update a object
            Args:
                id (int): Target Object ID
                newObj (object): New object will be merged into original object
            Returns:
                Object: Updated object
                None: If specified object id is not found
                MultipleInvalid: If input object is invaild
        """
        newObj = self.validation(newObj)
        for obj in self.model.db:
            if obj["id"] != id:
                continue

            newObj.pop("id", None)
            obj.update(newObj)
            obj = self._cast_model(obj)
            if not self._batch.enable.is_set():
                self.model.save_db()
            return obj

        return None

    def set(self, id, newObj):
        """Set a object
            Args:
                id (int): Target Object ID
                newObj (object): New object will be set
            Returns:
                Object: New object
                None: If specified object id is not found
                MultipleInvalid: If input object is invaild
        """
        newObj = self.validation(newObj)
        for index in moves.range(0, len(self.model.db)):
            if self.model.db[index]["id"] != id:
                continue

            newObj["id"] = id
            self.model.db[index] = self._cast_model(newObj)
            if not self._batch.enable.is_set():
                self.model.save_db()
            return self.model.db[index]

        return None

    def getAll(self):
        """Get all objects
            Returns:
                List: list of all objects
        """
        objs = []
        for obj in self.model.db:
            objs.append(self._cast_model(obj))

        return objs



from __future__ import print_function

import os
import sys
import simplejson as json
import uuid
from mock import Mock

from random import randint
from threading import Thread
from threading import Lock
from threading import Event

try:
    from queue import Queue
except ImportError:
    from Queue import Queue

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + '/../../')
    from sanji.connection.connection import Connection
except ImportError as e:
    print(e)
    print("Please check the python PATH for import test module. (%s)"
          % __file__)
    exit(1)


class Object(object):
    pass


class Mockup(Connection):

    def __init__(self):
        self.message_queue = Queue()
        self.publish_onfly = {}
        self.tunnels = {
            "internel": (uuid.uuid4().hex, None),
            "model": (None, None),
            "view": (None, None)
        }
        self.disconnect_event = Event()
        self._publish_lock = Lock()
        self.on_publish = None
        self.on_message = None
        self.on_connect = lambda client, userdata, flags, rc: 0
        self.message_callback_add = Mock()
        self.message_callback_remove = Mock()

    def __onpublish(self):
        while self.disconnect_event.is_set() is False:
            arrival = []
            self._publish_lock.acquire()
            for mid in self.publish_onfly:
                arrival.append(mid)
            self.publish_onfly = {}
            self._publish_lock.release()

            def pub(mid):
                self.on_publish(self, None, mid)

            map(pub, arrival)

    def __onmessage(self):
        while self.disconnect_event.is_set() is False:
            while self.message_queue.empty() is False:
                message = self.message_queue.get()
                message = json.loads(message)
                msg_obj = Object()
                for key in message:
                    msg_obj.__setattr__(key, message[key])
                self.on_message(self, None, msg_obj)

    def connect(self):
        self.__t_onpublish = Thread(target=self.__onpublish)
        self.__t_onpublish.daemon = True
        self.__t_onpublish.start()

        self.__t_onmessage = Thread(target=self.__onmessage)
        self.__t_onmessage.daemon = True
        self.__t_onmessage.start()

        self.on_connect(self, None, None, 0)

        return 0

    def disconnect(self):
        self.disconnect_event.set()
        return 0

    def set_tunnels(self, tunnels):
        self.tunnels = tunnels
        return tunnels

    def set_on_connect(self, func):
        self.on_connect = func

    def set_on_message(self, func):
        self.on_message = func

    def set_on_publish(self, func):
        self.on_publish = func

    def set_on_connection(self, func):
        self.on_connect = func

    def publish(self, **kwargs):
        # self.message_queue.put(json.dumps(kwargs))
        mid = randint(0, 65535)
        self._publish_lock.acquire()
        self.publish_onfly[mid] = kwargs
        self._publish_lock.release()

        return mid



#!/usr/bin/env python
# -*- coding: UTF-8 -*-


import logging
import uuid
import six

from collections import deque
from threading import Event
from threading import RLock
from threading import Thread
from time import sleep
from time import time

from sanji.message import Message


_logger = logging.getLogger("sanji.sdk.session")


class Status(object):
    """
    Status of session
    """
    CREATED = 0
    SENDING = 1
    SENT = 2
    RESOLVED = 3
    SEND_TIMEOUT = 4
    RESPONSE_TIMEOUT = 4


class TimeoutError(Exception):
    pass


class StatusError(Exception):
    pass


class SessionError(Exception):
    pass


class Session(object):
    """
    Session
    """
    def __init__(self):
        self.aging_unit = 0.5
        self.session_list = {}
        self.session_lock = RLock()
        self.timeout_queue = deque([], maxlen=10)
        self.stop_event = Event()
        self.thread_aging = Thread(target=self.aging)
        self.thread_aging.daemon = True
        self.thread_aging.start()

    def stop(self):
        self.stop_event.set()
        if self.thread_aging.is_alive():
            self.thread_aging.join()

    def resolve(self, msg_id, message=None, status=Status.RESOLVED):
        with self.session_lock:
            session = self.session_list.pop(msg_id, None)
            if session is None:
                return

            session["resolve_message"] = message
            session["status"] = status
            session["is_resolved"].set()
            return session

    def resolve_send(self, mid_id):
        with self.session_lock:
            for session in six.itervalues(self.session_list):
                if session["mid"] == mid_id:
                    session["status"] = Status.SENT
                    session["is_published"].set()
                    return session
            _logger.debug("Nothing can be resolved mid_id: %s" % mid_id)
            return None

    def create(self, message, mid=None, age=60, force=True):
        """
        create session
            force if you pass `force = False`, it may raise SessionError
                due to duplicate message id
        """
        with self.session_lock:
            if not hasattr(message, "id"):
                message.__setattr__("id", "event-%s" % (uuid.uuid4().hex,))
            if self.session_list.get(message.id, None) is not None:
                if force is False:
                    raise SessionError("Message id: %s duplicate!" %
                                       message.id)
                else:
                    message = Message(message.to_dict(), generate_id=True)

            session = {
                "status": Status.CREATED,
                "message": message,
                "age": age,
                "mid": mid,
                "created_at": time(),
                "is_published": Event(),
                "is_resolved": Event()
            }
            self.session_list.update({
                message.id: session
            })

            return session

    def aging(self):
        while not self.stop_event.is_set():
            with self.session_lock:
                for session_id in self.session_list:
                    session = self.session_list[session_id]
                    # TODO: use system time diff to decrease age
                    #       instead of just - 1 ?
                    session["age"] = session["age"] - self.aging_unit
                    # age > 0
                    if session["age"] > 0:
                        continue

                    # age <= 0, timeout!
                    _logger.debug("Message timeout id:%s", session_id)
                    if session["is_published"].is_set():
                        session["status"] = Status.SEND_TIMEOUT
                    else:
                        session["status"] = Status.RESPONSE_TIMEOUT
                    session["is_published"].set()
                    session["is_resolved"].set()

                    self.timeout_queue.append(session)

                # remove all timeout session
                self.session_list = dict((k, self.session_list[k]) for k
                                         in self.session_list
                                         if self.session_list[k]["status"] !=
                                         Status.SEND_TIMEOUT or
                                         self.session_list[k]["status"] !=
                                         Status.RESPONSE_TIMEOUT)
            sleep(self.aging_unit)



from setuptools import setup
import os


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()


setup(
    name="sanji",
    version="1.1.0",
    description="Sanji Framework SDK",
    long_description=read('README.rst'),
    url="https://github.com/Sanji-IO/sanji",
    author="Sanji Team",
    author_email="sanji@moxa.com",
    license="MIT",
    packages=["sanji", "sanji.connection", "sanji.model"],
    install_requires=["voluptuous", "simplejson"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
